@@ -1,7 +1,7 @@
 /* StarPU --- Runtime system for heterogeneous multicore architectures.
  *
- * Copyright (C) 2009-2010, 2012-2013  Université de Bordeaux 1
- * Copyright (C) 2010, 2011, 2012, 2013  Centre National de la Recherche Scientifique
+ * Copyright (C) 2009-2010, 2012-2016  Université de Bordeaux
+ * Copyright (C) 2010, 2011, 2012, 2013, 2014, 2015  CNRS
  *
  * StarPU is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -18,12 +18,38 @@
 #include <errno.h>
 
 #include <core/workers.h>
+#include <core/disk.h>
 #include <common/config.h>
+#include <common/fxt.h>
 #include <starpu.h>
 #include <drivers/opencl/driver_opencl.h>
 #include <datawizard/memory_manager.h>
+#include <datawizard/memory_nodes.h>
+#include <datawizard/malloc.h>
+#include <core/simgrid.h>
+
+#ifdef STARPU_SIMGRID
+#include <sys/mman.h>
+#include <fcntl.h>
+#endif
+
+#ifndef O_BINARY
+#define O_BINARY 0
+#endif
+
+#ifndef MAP_POPULATE
+#define MAP_POPULATE 0
+#endif
 
 static size_t _malloc_align = sizeof(void*);
+static int disable_pinning;
+static int malloc_on_node_default_flags[STARPU_MAXNODES];
+
+/* This file is used for implementing "folded" allocation */
+#ifdef STARPU_SIMGRID
+static int bogusfile = -1;
+static unsigned long _starpu_malloc_simulation_fold;
+#endif
 
 void starpu_malloc_set_align(size_t align)
 {
@@ -47,7 +73,7 @@ struct malloc_pinned_codelet_struct
 //{
 //	struct malloc_pinned_codelet_struct *s = arg;
 //        //        *(s->ptr) = malloc(s->dim);
-//        starpu_opencl_allocate_memory((void **)(s->ptr), s->dim, CL_MEM_READ_WRITE|CL_MEM_ALLOC_HOST_PTR);
+//        starpu_opencl_allocate_memory(devid, (void **)(s->ptr), s->dim, CL_MEM_READ_WRITE|CL_MEM_ALLOC_HOST_PTR);
 //}
 //#endif
 
@@ -72,9 +98,9 @@ static struct starpu_perfmodel malloc_pinned_model =
 
 static struct starpu_codelet malloc_pinned_cl =
 {
-	.cuda_funcs = {malloc_pinned_cuda_codelet, NULL},
+	.cuda_funcs = {malloc_pinned_cuda_codelet},
 //#ifdef STARPU_USE_OPENCL
-//	.opencl_funcs = {malloc_pinned_opencl_codelet, NULL},
+//	.opencl_funcs = {malloc_pinned_opencl_codelet},
 //#endif
 	.nbuffers = 0,
 	.model = &malloc_pinned_model
@@ -89,31 +115,37 @@ int starpu_malloc_flags(void **A, size_t dim, int flags)
 
 	if (flags & STARPU_MALLOC_COUNT)
 	{
-		if (_starpu_memory_manager_can_allocate_size(dim, 0) == 0)
-		{
-			size_t freed;
-			size_t reclaim = 2 * dim;
-			_STARPU_DEBUG("There is not enough memory left, we are going to reclaim %ld\n", reclaim);
-			_STARPU_TRACE_START_MEMRECLAIM(0);
-			freed = _starpu_memory_reclaim_generic(0, 0, reclaim);
-			_STARPU_TRACE_END_MEMRECLAIM(0);
-			if (freed < dim)
+		if (!(flags & STARPU_MALLOC_NORECLAIM))
+			while (starpu_memory_allocate(STARPU_MAIN_RAM, dim, flags) != 0)
 			{
-				// We could not reclaim enough memory
-				*A = NULL;
-				return -ENOMEM;
+				size_t freed;
+				size_t reclaim = 2 * dim;
+				_STARPU_DEBUG("There is not enough memory left, we are going to reclaim %ld\n", reclaim);
+				_STARPU_TRACE_START_MEMRECLAIM(STARPU_MAIN_RAM,0);
+				freed = _starpu_memory_reclaim_generic(STARPU_MAIN_RAM, 0, reclaim);
+				_STARPU_TRACE_END_MEMRECLAIM(STARPU_MAIN_RAM,0);
+				if (freed < dim && !(flags & STARPU_MEMORY_WAIT))
+				{
+					// We could not reclaim enough memory
+					*A = NULL;
+					return -ENOMEM;
+				}
 			}
-		}
+		else if (flags & STARPU_MEMORY_WAIT)
+			starpu_memory_allocate(STARPU_MAIN_RAM, dim, flags);
+		else
+			starpu_memory_allocate(STARPU_MAIN_RAM, dim, flags | STARPU_MEMORY_OVERFLOW);
 	}
 
-	if (flags & STARPU_MALLOC_PINNED)
+	if (flags & STARPU_MALLOC_PINNED && disable_pinning <= 0 && STARPU_RUNNING_ON_VALGRIND == 0)
 	{
 #ifdef STARPU_SIMGRID
 		/* FIXME: CUDA seems to be taking 650µs every 1MiB.
 		 * Ideally we would simulate this batching in 1MiB requests
 		 * instead of computing an average value.
 		 */
-		MSG_process_sleep((float) dim * 0.000650 / 1048576.);
+		if (_starpu_simgrid_cuda_malloc_cost())
+			MSG_process_sleep((float) dim * 0.000650 / 1048576.);
 #else /* STARPU_SIMGRID */
 		if (_starpu_can_submit_cuda_task())
 		{
@@ -122,13 +154,17 @@ int starpu_malloc_flags(void **A, size_t dim, int flags)
 			cudaError_t cures;
 			cures = cudaHostAlloc(A, dim, cudaHostAllocPortable);
 			if (STARPU_UNLIKELY(cures))
+			{
 				STARPU_CUDA_REPORT_ERROR(cures);
+				ret = -ENOMEM;
+			}
 			goto end;
 #else
 			int push_res;
 
-			if (STARPU_UNLIKELY(!_starpu_worker_may_perform_blocking_calls()))
-				return -EDEADLK;
+			/* Old versions of CUDA are not thread-safe, we have to
+			 * run cudaHostAlloc from CUDA workers */
+			STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "without CUDA peer allocation support, pinned allocation must not be done from task or callback");
 
 			struct malloc_pinned_codelet_struct s =
 			{
@@ -138,6 +174,7 @@ int starpu_malloc_flags(void **A, size_t dim, int flags)
 
 			malloc_pinned_cl.where = STARPU_CUDA;
 			struct starpu_task *task = starpu_task_create();
+			task->name = "cuda_malloc_pinned";
 			task->callback_func = NULL;
 			task->cl = &malloc_pinned_cl;
 			task->cl_arg = &s;
@@ -157,8 +194,7 @@ int starpu_malloc_flags(void **A, size_t dim, int flags)
 //#ifdef STARPU_USE_OPENCL
 //			int push_res;
 //
-//			if (STARPU_UNLIKELY(!_starpu_worker_may_perform_blocking_calls()))
-//				return -EDEADLK;
+//			STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "pinned OpenCL allocation must not be done from task or callback");
 //
 //			struct malloc_pinned_codelet_struct s =
 //				{
@@ -168,6 +204,7 @@ int starpu_malloc_flags(void **A, size_t dim, int flags)
 //
 //			malloc_pinned_cl.where = STARPU_OPENCL;
 //			struct starpu_task *task = starpu_task_create();
+//		        task->name = "opencl_malloc_pinned";
 //			task->callback_func = NULL;
 //			task->cl = &malloc_pinned_cl;
 //			task->cl_arg = &s;
@@ -183,6 +220,71 @@ int starpu_malloc_flags(void **A, size_t dim, int flags)
 #endif /* STARPU_SIMGRID */
 	}
 
+#ifdef STARPU_SIMGRID
+	if (flags & STARPU_MALLOC_SIMULATION_FOLDED)
+	{
+		/* Use "folded" allocation: the same file is mapped several
+		 * times contiguously, to get a memory area one can read/write,
+		 * without consuming memory */
+
+		/* First reserve memory area */
+		void *buf = mmap (NULL, dim, PROT_READ|PROT_WRITE, MAP_ANONYMOUS|MAP_PRIVATE, -1, 0);
+		unsigned i;
+		if (buf == MAP_FAILED)
+		{
+			_STARPU_DISP("Warning: could not allocate %luMiB of memory, you need to run \"sysctl vm.overcommit_memory=1\" as root to allow so big allocations\n", (unsigned long) (dim >> 20));
+			ret = -ENOMEM;
+			*A = NULL;
+		}
+		else
+		{
+			if (bogusfile == -1)
+			{
+				char *path = starpu_getenv("TMPDIR");
+				if (!path)
+					path = "/tmp";
+				/* Create bogus file if not done already */
+				char *name = _starpu_mktemp(path, O_RDWR | O_BINARY, &bogusfile);
+				char *dumb;
+				if (!name)
+				{
+					ret = errno;
+					munmap(buf, dim);
+					*A = NULL;
+					goto end;
+				}
+				unlink(name);
+				free(name);
+				dumb = calloc(1,_starpu_malloc_simulation_fold);
+				write(bogusfile, dumb, _starpu_malloc_simulation_fold);
+				free(dumb);
+			}
+			/* Map the bogus file in place of the anonymous memory */
+			for (i = 0; i < dim / _starpu_malloc_simulation_fold; i++)
+			{
+				void *pos = (void*) ((unsigned long) buf + i * _starpu_malloc_simulation_fold);
+				void *res = mmap(pos, _starpu_malloc_simulation_fold, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_SHARED|MAP_POPULATE, bogusfile, 0);
+				STARPU_ASSERT_MSG(res == pos, "Could not map folded virtual memory (%s). Do you perhaps need to increase the STARPU_MALLOC_SIMULATION_FOLD environment variable or the sysctl vm.max_map_count?", strerror(errno));
+			}
+
+			if (dim % _starpu_malloc_simulation_fold)
+			{
+				void *pos = (void*) ((unsigned long) buf + i * _starpu_malloc_simulation_fold);
+				void *res = mmap(pos, dim % _starpu_malloc_simulation_fold, PROT_READ|PROT_WRITE, MAP_FIXED|MAP_SHARED|MAP_POPULATE, bogusfile, 0);
+				STARPU_ASSERT_MSG(res == pos, "Could not map folded virtual memory (%s). Do you perhaps need to increase the STARPU_MALLOC_SIMULATION_FOLD environment variable or the sysctl vm.max_map_count?", strerror(errno));
+			}
+			*A = buf;
+		}
+	}
+	else
+#endif
+	if (_starpu_can_submit_scc_task())
+	{
+#ifdef STARPU_USE_SCC
+		_starpu_scc_allocate_shared_memory(A, dim);
+#endif
+	}
+	else
 #ifdef STARPU_HAVE_POSIX_MEMALIGN
 	if (_malloc_align != sizeof(void*))
 	{
@@ -197,17 +299,27 @@ int starpu_malloc_flags(void **A, size_t dim, int flags)
 		if (_malloc_align != sizeof(void*))
 		{
 			*A = memalign(_malloc_align, dim);
+			if (!*A)
+				ret = -ENOMEM;
 		}
 		else
 #endif /* STARPU_HAVE_POSIX_MEMALIGN */
 		{
 			*A = malloc(dim);
+			if (!*A)
+				ret = -ENOMEM;
 		}
 
+#if defined(STARPU_SIMGRID) || defined(STARPU_USE_CUDA)
 end:
+#endif
 	if (ret == 0)
 	{
-		STARPU_ASSERT(*A);
+		STARPU_ASSERT_MSG(*A, "Failed to allocated memory of size %ld b\n", dim);
+	}
+	else if (flags & STARPU_MALLOC_COUNT)
+	{
+		starpu_memory_deallocate(STARPU_MAIN_RAM, dim);
 	}
 
 	return ret;
@@ -246,9 +358,9 @@ static struct starpu_perfmodel free_pinned_model =
 
 static struct starpu_codelet free_pinned_cl =
 {
-	.cuda_funcs = {free_pinned_cuda_codelet, NULL},
+	.cuda_funcs = {free_pinned_cuda_codelet},
 //#ifdef STARPU_USE_OPENCL
-//	.opencl_funcs = {free_pinned_opencl_codelet, NULL},
+//	.opencl_funcs = {free_pinned_opencl_codelet},
 //#endif
 	.nbuffers = 0,
 	.model = &free_pinned_model
@@ -258,7 +370,7 @@ static struct starpu_codelet free_pinned_cl =
 int starpu_free_flags(void *A, size_t dim, int flags)
 {
 #ifndef STARPU_SIMGRID
-	if (flags & STARPU_MALLOC_PINNED)
+	if (flags & STARPU_MALLOC_PINNED && disable_pinning <= 0 && STARPU_RUNNING_ON_VALGRIND == 0)
 	{
 		if (_starpu_can_submit_cuda_task())
 		{
@@ -280,11 +392,11 @@ int starpu_free_flags(void *A, size_t dim, int flags)
 			{
 				int push_res;
 
-				if (STARPU_UNLIKELY(!_starpu_worker_may_perform_blocking_calls()))
-					return -EDEADLK;
+				STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "without CUDA peer allocation support, pinned deallocation must not be done from task or callback");
 
 				free_pinned_cl.where = STARPU_CUDA;
 				struct starpu_task *task = starpu_task_create();
+				task->name = "cuda_free_pinned";
 				task->callback_func = NULL;
 				task->cl = &free_pinned_cl;
 				task->cl_arg = A;
@@ -304,11 +416,11 @@ int starpu_free_flags(void *A, size_t dim, int flags)
 //#ifdef STARPU_USE_OPENCL
 //		int push_res;
 //
-//		if (STARPU_UNLIKELY(!_starpu_worker_may_perform_blocking_calls()))
-//			return -EDEADLK;
+//		STARPU_ASSERT_MSG(_starpu_worker_may_perform_blocking_calls(), "pinned OpenCL deallocation must not be done from task or callback");
 //
 //                free_pinned_cl.where = STARPU_OPENCL;
 //		struct starpu_task *task = starpu_task_create();
+//              task->name = "opencl_free_pinned";
 //		task->callback_func = NULL;
 //		task->cl = &free_pinned_cl;
 //		task->cl_arg = A;
@@ -324,12 +436,28 @@ int starpu_free_flags(void *A, size_t dim, int flags)
 	}
 #endif /* STARPU_SIMGRID */
 
-	free(A);
+#ifdef STARPU_SIMGRID
+	if (flags & STARPU_MALLOC_SIMULATION_FOLDED)
+	{
+		munmap(A, dim);
+	}
+	else
+#endif
+	if (_starpu_can_submit_scc_task())
+	{
+#ifdef STARPU_USE_SCC
+		_starpu_scc_free_shared_memory(A);
+#endif
+	}
+	else
+		free(A);
 
+#if !defined(STARPU_SIMGRID) && defined(STARPU_USE_CUDA)
 out:
+#endif
 	if (flags & STARPU_MALLOC_COUNT)
 	{
-		_starpu_memory_manager_deallocate_size(dim, 0);
+		starpu_memory_deallocate(STARPU_MAIN_RAM, dim);
 	}
 
 	return 0;
@@ -346,22 +474,39 @@ static starpu_pthread_mutex_t opencl_alloc_mutex = STARPU_PTHREAD_MUTEX_INITIALI
 #endif
 
 static uintptr_t
-_starpu_malloc_on_node(unsigned dst_node, size_t size)
+_starpu_malloc_on_node(unsigned dst_node, size_t size, int flags)
 {
 	uintptr_t addr = 0;
 
-#ifdef STARPU_USE_CUDA
+#if defined(STARPU_USE_CUDA) && !defined(STARPU_SIMGRID)
 	cudaError_t status;
 #endif
 
-	if (_starpu_memory_manager_can_allocate_size(size, dst_node) == 0)
-		return 0;
+	/* Handle count first */
+	if (flags & STARPU_MALLOC_COUNT)
+	{
+		if (starpu_memory_allocate(dst_node, size, flags) != 0)
+			return 0;
+		/* And prevent double-count in starpu_malloc_flags */
+		flags &= ~STARPU_MALLOC_COUNT;
+	}
 
 	switch(starpu_node_get_kind(dst_node))
 	{
 		case STARPU_CPU_RAM:
 		{
-			addr = (uintptr_t)malloc(size);
+			starpu_malloc_flags((void**) &addr, size,
+#if defined(STARPU_USE_CUDA) && !defined(HAVE_CUDA_MEMCPY_PEER) && !defined(STARPU_SIMGRID)
+					/* without memcpy_peer, we can not
+					 * allocated pinned memory, since it
+					 * requires waiting for a task, and we
+					 * may be called with a spinlock held
+					 */
+					flags & ~STARPU_MALLOC_PINNED
+#else
+					flags
+#endif
+					);
 			break;
 		}
 #if defined(STARPU_USE_CUDA) || defined(STARPU_SIMGRID)
@@ -374,7 +519,8 @@ _starpu_malloc_on_node(unsigned dst_node, size_t size)
 #endif
 			/* Sleep for the allocation */
 			STARPU_PTHREAD_MUTEX_LOCK(&cuda_alloc_mutex);
-			MSG_process_sleep(0.000175);
+			if (_starpu_simgrid_cuda_malloc_cost())
+				MSG_process_sleep(0.000175);
 			if (!last[dst_node])
 				last[dst_node] = 1<<10;
 			addr = last[dst_node];
@@ -382,6 +528,14 @@ _starpu_malloc_on_node(unsigned dst_node, size_t size)
 			STARPU_ASSERT(last[dst_node] >= addr);
 			STARPU_PTHREAD_MUTEX_UNLOCK(&cuda_alloc_mutex);
 #else
+			struct _starpu_worker *worker = _starpu_get_local_worker_key();
+			unsigned devid = _starpu_memory_node_get_devid(dst_node);
+			if (!worker || worker->arch != STARPU_CUDA_WORKER || worker->devid != devid)
+#if defined(HAVE_CUDA_MEMCPY_PEER)
+				starpu_cuda_set_device(devid);
+#else
+				STARPU_ASSERT_MSG(0, "CUDA peer access is not available with this version of CUDA");
+#endif
 			status = cudaMalloc((void **)&addr, size);
 			if (!addr || (status != cudaSuccess))
 			{
@@ -400,7 +554,8 @@ _starpu_malloc_on_node(unsigned dst_node, size_t size)
 				static uintptr_t last[STARPU_MAXNODES];
 				/* Sleep for the allocation */
 				STARPU_PTHREAD_MUTEX_LOCK(&opencl_alloc_mutex);
-				MSG_process_sleep(0.000175);
+				if (_starpu_simgrid_cuda_malloc_cost())
+					MSG_process_sleep(0.000175);
 				if (!last[dst_node])
 					last[dst_node] = 1<<10;
 				addr = last[dst_node];
@@ -411,7 +566,7 @@ _starpu_malloc_on_node(unsigned dst_node, size_t size)
                                 int ret;
 				cl_mem ptr;
 
-				ret = starpu_opencl_allocate_memory(&ptr, size, CL_MEM_READ_WRITE);
+				ret = starpu_opencl_allocate_memory(_starpu_memory_node_get_devid(dst_node), &ptr, size, CL_MEM_READ_WRITE);
 				if (ret)
 				{
 					addr = 0;
@@ -423,6 +578,24 @@ _starpu_malloc_on_node(unsigned dst_node, size_t size)
 				break;
 #endif
 		}
+#endif
+	        case STARPU_DISK_RAM:
+		{
+			addr = (uintptr_t) _starpu_disk_alloc(dst_node, size);
+			break;
+		}
+
+#ifdef STARPU_USE_MIC
+		case STARPU_MIC_RAM:
+			if (_starpu_mic_allocate_memory((void **)(&addr), size, dst_node))
+				addr = 0;
+			break;
+#endif
+#ifdef STARPU_USE_SCC
+		case STARPU_SCC_RAM:
+			if (_starpu_scc_allocate_memory((void **)(&addr), size, dst_node))
+				addr = 0;
+			break;
 #endif
 		default:
 			STARPU_ABORT();
@@ -431,19 +604,28 @@ _starpu_malloc_on_node(unsigned dst_node, size_t size)
 	if (addr == 0)
 	{
 		// Allocation failed, gives the memory back to the memory manager
-		_starpu_memory_manager_deallocate_size(size, dst_node);
+		_STARPU_TRACE_MEMORY_FULL(size);
+		starpu_memory_deallocate(dst_node, size);
 	}
 	return addr;
 }
 
 void
-_starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
+_starpu_free_on_node_flags(unsigned dst_node, uintptr_t addr, size_t size, int flags)
 {
+	int count = flags & STARPU_MALLOC_COUNT;
+	flags &= ~STARPU_MALLOC_COUNT;
 	enum starpu_node_kind kind = starpu_node_get_kind(dst_node);
 	switch(kind)
 	{
 		case STARPU_CPU_RAM:
-			free((void*)addr);
+			starpu_free_flags((void*)addr, size,
+#if defined(STARPU_USE_CUDA) && !defined(HAVE_CUDA_MEMCPY_PEER) && !defined(STARPU_SIMGRID)
+					flags & ~STARPU_MALLOC_PINNED
+#else
+					flags
+#endif
+					);
 			break;
 #if defined(STARPU_USE_CUDA) || defined(STARPU_SIMGRID)
 		case STARPU_CUDA_RAM:
@@ -451,12 +633,31 @@ _starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
 #ifdef STARPU_SIMGRID
 			STARPU_PTHREAD_MUTEX_LOCK(&cuda_alloc_mutex);
 			/* Sleep for the free */
-			MSG_process_sleep(0.000750);
+			if (_starpu_simgrid_cuda_malloc_cost())
+				MSG_process_sleep(0.000750);
 			STARPU_PTHREAD_MUTEX_UNLOCK(&cuda_alloc_mutex);
 #else
 			cudaError_t err;
+			struct _starpu_worker *worker = _starpu_get_local_worker_key();
+			unsigned devid = _starpu_memory_node_get_devid(dst_node);
+			if (!worker || worker->arch != STARPU_CUDA_WORKER || worker->devid != devid)
+#if defined(HAVE_CUDA_MEMCPY_PEER)
+				starpu_cuda_set_device(devid);
+#else
+				STARPU_ASSERT_MSG(0, "CUDA peer access is not available with this version of CUDA");
+#endif
 			err = cudaFree((void*)addr);
-			if (STARPU_UNLIKELY(err != cudaSuccess))
+			if (STARPU_UNLIKELY(err != cudaSuccess
+#ifdef STARPU_OPENMP
+		/* When StarPU is used as Open Runtime support,
+		 * starpu_omp_shutdown() will usually be called from a
+		 * destructor, in which case cudaThreadExit() reports a
+		 * cudaErrorCudartUnloading here. There should not
+		 * be any remaining tasks running at this point so
+		 * we can probably ignore it without much consequences. */
+		&& err != cudaErrorCudartUnloading
+#endif /* STARPU_OPENMP */
+						))
 				STARPU_CUDA_REPORT_ERROR(err);
 #endif
 			break;
@@ -468,7 +669,8 @@ _starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
 #ifdef STARPU_SIMGRID
 			STARPU_PTHREAD_MUTEX_LOCK(&opencl_alloc_mutex);
 			/* Sleep for the free */
-			MSG_process_sleep(0.000750);
+			if (_starpu_simgrid_cuda_malloc_cost())
+				MSG_process_sleep(0.000750);
 			STARPU_PTHREAD_MUTEX_UNLOCK(&opencl_alloc_mutex);
 #else
 			cl_int err;
@@ -478,14 +680,57 @@ _starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
 #endif
                         break;
 		}
+#endif
+	        case STARPU_DISK_RAM:
+		{
+			_starpu_disk_free (dst_node, (void *) addr , size);
+			break;
+		}
+
+#ifdef STARPU_USE_MIC
+		case STARPU_MIC_RAM:
+			_starpu_mic_free_memory((void*) addr, size, dst_node);
+			break;
+#endif
+#ifdef STARPU_USE_SCC
+		case STARPU_SCC_RAM:
+			_starpu_scc_free_memory((void *) addr, dst_node);
+			break;
 #endif
 		default:
 			STARPU_ABORT();
 	}
-	_starpu_memory_manager_deallocate_size(size, dst_node);
+	if (count)
+		starpu_memory_deallocate(dst_node, size);
 
 }
 
+int
+starpu_memory_pin(void *addr STARPU_ATTRIBUTE_UNUSED, size_t size STARPU_ATTRIBUTE_UNUSED)
+{
+	if (STARPU_MALLOC_PINNED && disable_pinning <= 0 && STARPU_RUNNING_ON_VALGRIND == 0)
+	{
+#if defined(STARPU_USE_CUDA) && defined(HAVE_CUDA_MEMCPY_PEER)
+		if (cudaHostRegister(addr, size, cudaHostRegisterPortable) != cudaSuccess)
+			return -1;
+#endif
+	}
+	return 0;
+}
+
+int
+starpu_memory_unpin(void *addr STARPU_ATTRIBUTE_UNUSED, size_t size STARPU_ATTRIBUTE_UNUSED)
+{
+	if (STARPU_MALLOC_PINNED && disable_pinning <= 0 && STARPU_RUNNING_ON_VALGRIND == 0)
+	{
+#if defined(STARPU_USE_CUDA) && defined(HAVE_CUDA_MEMCPY_PEER)
+		if (cudaHostUnregister(addr) != cudaSuccess)
+			return -1;
+#endif
+	}
+	return 0;
+}
+
 /*
  * On CUDA which has very expensive malloc, for small sizes, allocate big
  * chunks divided in blocks, and we actually allocate segments of consecutive
@@ -512,7 +757,8 @@ _starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
 #define CHUNK_NBLOCKS (CHUNK_SIZE/CHUNK_ALLOC_MIN)
 
 /* Linked list for available segments */
-struct block {
+struct block
+{
 	int length;	/* Number of consecutive free blocks */
 	int next;	/* next free segment */
 };
@@ -533,7 +779,7 @@ LIST_TYPE(_starpu_chunk,
 )
 
 /* One list of chunks per node */
-static struct _starpu_chunk_list *chunks[STARPU_MAXNODES];
+static struct _starpu_chunk_list chunks[STARPU_MAXNODES];
 /* Number of completely free chunks */
 static int nfreechunks[STARPU_MAXNODES];
 /* This protects chunks and nfreechunks */
@@ -542,9 +788,15 @@ static starpu_pthread_mutex_t chunk_mutex[STARPU_MAXNODES];
 void
 _starpu_malloc_init(unsigned dst_node)
 {
-	chunks[dst_node] = _starpu_chunk_list_new();
+	_starpu_chunk_list_init(&chunks[dst_node]);
 	nfreechunks[dst_node] = 0;
 	STARPU_PTHREAD_MUTEX_INIT(&chunk_mutex[dst_node], NULL);
+	disable_pinning = starpu_get_env_number("STARPU_DISABLE_PINNING");
+	malloc_on_node_default_flags[dst_node] = STARPU_MALLOC_PINNED | STARPU_MALLOC_COUNT;
+#ifdef STARPU_SIMGRID
+	/* Reasonably "costless" */
+	_starpu_malloc_simulation_fold = starpu_get_env_number_default("STARPU_MALLOC_SIMULATION_FOLD", 1) << 20;
+#endif
 }
 
 void
@@ -552,30 +804,25 @@ _starpu_malloc_shutdown(unsigned dst_node)
 {
 	struct _starpu_chunk *chunk, *next_chunk;
 
-	if (!chunks[dst_node])
-		return;
-
 	STARPU_PTHREAD_MUTEX_LOCK(&chunk_mutex[dst_node]);
-	for (chunk = _starpu_chunk_list_begin(chunks[dst_node]);
-	     chunk != _starpu_chunk_list_end(chunks[dst_node]);
+	for (chunk = _starpu_chunk_list_begin(&chunks[dst_node]);
+	     chunk != _starpu_chunk_list_end(&chunks[dst_node]);
 	     chunk = next_chunk)
 	{
 		next_chunk = _starpu_chunk_list_next(chunk);
-		_starpu_free_on_node(dst_node, chunk->base, CHUNK_SIZE);
-		_starpu_chunk_list_erase(chunks[dst_node], chunk);
+		_starpu_free_on_node_flags(dst_node, chunk->base, CHUNK_SIZE, malloc_on_node_default_flags[dst_node]);
+		_starpu_chunk_list_erase(&chunks[dst_node], chunk);
 		free(chunk);
 	}
-	_starpu_chunk_list_delete(chunks[dst_node]);
-	chunks[dst_node] = NULL;
 	STARPU_PTHREAD_MUTEX_UNLOCK(&chunk_mutex[dst_node]);
 	STARPU_PTHREAD_MUTEX_DESTROY(&chunk_mutex[dst_node]);
 }
 
 /* Create a new chunk */
-static struct _starpu_chunk *_starpu_new_chunk(unsigned dst_node)
+static struct _starpu_chunk *_starpu_new_chunk(unsigned dst_node, int flags)
 {
 	struct _starpu_chunk *chunk;
-	uintptr_t base = _starpu_malloc_on_node(dst_node, CHUNK_SIZE);
+	uintptr_t base = _starpu_malloc_on_node(dst_node, CHUNK_SIZE, flags);
 
 	if (!base)
 		return NULL;
@@ -598,11 +845,11 @@ static struct _starpu_chunk *_starpu_new_chunk(unsigned dst_node)
 }
 
 uintptr_t
-starpu_malloc_on_node(unsigned dst_node, size_t size)
+starpu_malloc_on_node_flags(unsigned dst_node, size_t size, int flags)
 {
 	/* Big allocation, allocate normally */
 	if (size > CHUNK_ALLOC_MAX || starpu_node_get_kind(dst_node) != STARPU_CUDA_RAM)
-		return _starpu_malloc_on_node(dst_node, size);
+		return _starpu_malloc_on_node(dst_node, size, flags);
 
 	/* Round up allocation to block size */
 	int nblocks = (size + CHUNK_ALLOC_MIN - 1) / CHUNK_ALLOC_MIN;
@@ -615,8 +862,8 @@ starpu_malloc_on_node(unsigned dst_node, size_t size)
 	STARPU_PTHREAD_MUTEX_LOCK(&chunk_mutex[dst_node]);
 
 	/* Try to find a big enough segment among the chunks */
-	for (chunk = _starpu_chunk_list_begin(chunks[dst_node]);
-	     chunk != _starpu_chunk_list_end(chunks[dst_node]);
+	for (chunk = _starpu_chunk_list_begin(&chunks[dst_node]);
+	     chunk != _starpu_chunk_list_end(&chunks[dst_node]);
 	     chunk = _starpu_chunk_list_next(chunk))
 	{
 		if (chunk->available_max < nblocks)
@@ -630,15 +877,16 @@ starpu_malloc_on_node(unsigned dst_node, size_t size)
 		{
 			STARPU_ASSERT(block >= 0 && block <= CHUNK_NBLOCKS);
 			int length = bitmap[block].length;
-			if (length >= nblocks) {
+			if (length >= nblocks)
+			{
 
 				if (length >= 2*nblocks)
 				{
 					/* This one this has quite some room,
 					 * put it front, to make finding it
 					 * easier next time. */
-					_starpu_chunk_list_erase(chunks[dst_node], chunk);
-					_starpu_chunk_list_push_front(chunks[dst_node], chunk);
+					_starpu_chunk_list_erase(&chunks[dst_node], chunk);
+					_starpu_chunk_list_push_front(&chunks[dst_node], chunk);
 				}
 				if (chunk->available == CHUNK_NBLOCKS)
 					/* This one was empty, it's not empty any more */
@@ -655,7 +903,7 @@ starpu_malloc_on_node(unsigned dst_node, size_t size)
 	}
 
 	/* Didn't find a big enough segment, create another chunk.  */
-	chunk = _starpu_new_chunk(dst_node);
+	chunk = _starpu_new_chunk(dst_node, flags);
 	if (!chunk)
 	{
 		/* Really no memory any more, fail */
@@ -665,7 +913,7 @@ starpu_malloc_on_node(unsigned dst_node, size_t size)
 	}
 
 	/* And make it easy to find. */
-	_starpu_chunk_list_push_front(chunks[dst_node], chunk);
+	_starpu_chunk_list_push_front(&chunks[dst_node], chunk);
 	bitmap = chunk->bitmap;
 	prevblock = 0;
 	block = 1;
@@ -695,12 +943,12 @@ found:
 }
 
 void
-starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
+starpu_free_on_node_flags(unsigned dst_node, uintptr_t addr, size_t size, int flags)
 {
 	/* Big allocation, deallocate normally */
 	if (size > CHUNK_ALLOC_MAX || starpu_node_get_kind(dst_node) != STARPU_CUDA_RAM)
 	{
-		_starpu_free_on_node(dst_node, addr, size);
+		_starpu_free_on_node_flags(dst_node, addr, size, flags);
 		return;
 	}
 
@@ -710,12 +958,12 @@ starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
 	int nblocks = (size + CHUNK_ALLOC_MIN - 1) / CHUNK_ALLOC_MIN;
 
 	STARPU_PTHREAD_MUTEX_LOCK(&chunk_mutex[dst_node]);
-	for (chunk = _starpu_chunk_list_begin(chunks[dst_node]);
-	     chunk != _starpu_chunk_list_end(chunks[dst_node]);
+	for (chunk = _starpu_chunk_list_begin(&chunks[dst_node]);
+	     chunk != _starpu_chunk_list_end(&chunks[dst_node]);
 	     chunk = _starpu_chunk_list_next(chunk))
 		if (addr >= chunk->base && addr < chunk->base + CHUNK_SIZE)
 			break;
-	STARPU_ASSERT(chunk != _starpu_chunk_list_end(chunks[dst_node]));
+	STARPU_ASSERT(chunk != _starpu_chunk_list_end(&chunks[dst_node]));
 
 	struct block *bitmap = chunk->bitmap;
 	int block = ((addr - chunk->base) / CHUNK_ALLOC_MIN) + 1, prevblock, nextblock;
@@ -767,20 +1015,40 @@ starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
 	{
 		/* This chunk is now empty, but avoid chunk free/alloc
 		 * ping-pong by keeping some of these.  */
-		if (nfreechunks[dst_node] >= 1) {
+		if (nfreechunks[dst_node] >= 1)
+		{
 			/* We already have free chunks, release this one */
-			_starpu_free_on_node(dst_node, chunk->base, CHUNK_SIZE);
-			_starpu_chunk_list_erase(chunks[dst_node], chunk);
+			_starpu_free_on_node_flags(dst_node, chunk->base, CHUNK_SIZE, flags);
+			_starpu_chunk_list_erase(&chunks[dst_node], chunk);
 			free(chunk);
-		} else
+		}
+		else
 			nfreechunks[dst_node]++;
 	}
 	else
 	{
 		/* Freed some room, put this first in chunks list */
-		_starpu_chunk_list_erase(chunks[dst_node], chunk);
-		_starpu_chunk_list_push_front(chunks[dst_node], chunk);
+		_starpu_chunk_list_erase(&chunks[dst_node], chunk);
+		_starpu_chunk_list_push_front(&chunks[dst_node], chunk);
 	}
 
 	STARPU_PTHREAD_MUTEX_UNLOCK(&chunk_mutex[dst_node]);
 }
+
+void starpu_malloc_on_node_set_default_flags(unsigned node, int flags)
+{
+	STARPU_ASSERT_MSG(node < STARPU_MAXNODES, "bogus node value %d given to starpu_malloc_on_node_set_default_flags\n", node);
+	malloc_on_node_default_flags[node] = flags;
+}
+
+uintptr_t
+starpu_malloc_on_node(unsigned dst_node, size_t size)
+{
+	return starpu_malloc_on_node_flags(dst_node, size, malloc_on_node_default_flags[dst_node]);
+}
+
+void
+starpu_free_on_node(unsigned dst_node, uintptr_t addr, size_t size)
+{
+	starpu_free_on_node_flags(dst_node, addr, size, malloc_on_node_default_flags[dst_node]);
+}
